import keyword

from data_types import DataType, ObjectField, is_primitive_python_type
from operations import Operation


def escape_escapes(s: str) -> str:
    return s.replace("\\", "\\\\")


def docstring_comment(lines: list[str]) -> list[str]:
    if len(lines) == 0:
        return []
    if len(lines) == 1:
        return [f'"""{escape_escapes(lines[0])}"""']
    result = ['"""' + escape_escapes(lines[0])]
    result += [escape_escapes(line) for line in lines[1:]]
    result += ['"""']
    return result


def indent(lines: list[str], level: int) -> list[str]:
    """Indent each line.

    :param lines: Lines of text to be indented
    :param level: Level of indent (each indent level is four spaces)
    :return: Same lines of text provided after each line is indented
    """
    if level == 0:
        return lines
    else:
        return ["    " * level + line if line else "" for line in lines]


def data_type(d_type: DataType) -> list[str]:
    """Generate Python code defining the provided data type.

    :param d_type: Parsed API data type to render into Python code
    :return: Lines of Python code defining the provided data type
    """
    docstring_lines = (
        docstring_comment(d_type.description.split("\n")) if d_type.description else []
    )
    lines = []

    if d_type.enum_values:
        if any(str(t) in keyword.kwlist for v, t in d_type.enum_values.items()):
            lines.append(f"{d_type.name} = {d_type.python_type}")
            docstring_lines = (
                d_type.description.split("\n") if d_type.description else []
            )
            docstring_lines += ["", "Acceptable values:"] + [
                "* " + str(v) for v in d_type.enum_values
            ]
            lines.extend(docstring_comment(docstring_lines))
        else:
            lines.append(f"class {d_type.name}({d_type.python_type}, Enum):")
            lines.extend(indent(docstring_lines, 1))
            if docstring_lines:
                lines.append("")

            lines.extend(
                indent([f'{t} = "{v}"' for v, t in d_type.enum_values.items()], 1)
            )
    elif is_primitive_python_type(d_type.python_type):
        lines.append(f"{d_type.name} = {d_type.python_type}")
        lines.extend(docstring_lines)
    elif d_type.python_type == "ImplicitDict":
        if any(f.api_name in keyword.kwlist for f in d_type.fields):
            lines.append(f"{d_type.name} = dict")
            if d_type.description:
                docstring_lines = d_type.description.split("\n")
                docstring_lines.append("")
            else:
                docstring_lines = []
            docstring_lines.append("Expected keys:")
            docstring_lines.extend("* " + f.api_name for f in d_type.fields)
            lines.extend(docstring_comment(docstring_lines))
            lines.append("")
        else:
            lines.append(f"class {d_type.name}(ImplicitDict):")
            lines.extend(indent(docstring_lines, 1))
            if docstring_lines:
                lines.append("")

            for field in d_type.fields:
                lines.extend(indent(_object_field(field), 1))
                lines.append("")
            if d_type.fields:
                lines.pop()
    else:
        lines.append(f"{d_type.name} = {d_type.python_type}")

    return lines


def _object_field(field: ObjectField) -> list[str]:
    """Generate an unindented definition of the provided field in Python code.

    :param field: Data type field to render into Python code
    :return: Lines of Python code defining the provided field
    """
    d_type = field.python_type if field.required else f"Optional[{field.python_type}]"
    if field.default is not None:
        if field.literal_default:
            default_suffix = f" = {field.default}"
        elif isinstance(field.default, str):
            default_suffix = f' = "{field.default}"'
        else:
            default_suffix = f" = {str(field.default)}"
    else:
        default_suffix = ""
    lines = [f"{field.api_name}: {d_type}{default_suffix}"]
    if field.description:
        lines.extend(docstring_comment(field.description.split("\n")))
    return lines


def header(d_types: list[DataType]) -> list[str]:
    lines = ["# This file is autogenerated; do not modify manually!", ""]
    lines.extend(["from __future__ import annotations", ""])

    lines.append("from enum import Enum")

    basic_types = ["Dict"]
    if any(d.python_type.startswith("List[") for d in d_types) or any(
        any(f.python_type.startswith("List[") for f in d.fields) for d in d_types
    ):
        basic_types.append("List")
    if any(any(not f.required for f in d.fields) for d in d_types):
        basic_types.append("Optional")
    lines.append("from typing import " + ", ".join(basic_types))
    lines.append("")

    lines.append("from uas_standards.utils import Operation")
    lines.append("")

    lines.append("from implicitdict import ImplicitDict")
    if any(("StringBasedDateTime" in d.python_type) for d in d_types) or any(
        any(("StringBasedDateTime" in f.python_type) for f in d.fields) for d in d_types
    ):
        lines[-1] = lines[-1] + ", StringBasedDateTime"

    return lines


def data_types(d_types: list[DataType], default_package: str) -> list[str]:
    already_defined = [kw for kw in keyword.kwlist]
    already_defined += [
        "int",
        "float",
        "complex",
        "str",
        "list",
        "tuple",
        "range",
        "bytes",
        "bytearray",
        "memoryview",
        "dict",
        "bool",
        "set",
        "frozenset",
        "StringBasedDateTime",
        "ImplicitDict",
    ]

    lines = []

    # Declare types in dependency order
    total_defined = 0
    n_defined = 1
    not_defined = []

    def _core_type(type_name):
        core_type = type_name
        while "[" in core_type:
            core_type = core_type[core_type.index("[") + 1 :]
        while "]" in core_type:
            core_type = core_type[0 : core_type.index("]")]
        return core_type

    while n_defined > 0:
        n_defined = 0
        for d_type in d_types:
            if d_type.name in already_defined:
                continue
            if _core_type(d_type.python_type) not in already_defined:
                continue
            if any(
                (_core_type(f.python_type) not in already_defined)
                for f in d_type.fields
            ):
                continue

            lines.extend(["", ""])
            lines.extend(data_type(d_type))
            already_defined.append(d_type.name)
            n_defined += 1
            total_defined += 1

        not_defined = [
            d_type for d_type in d_types if d_type.name not in already_defined
        ]
        if not not_defined:
            break

        # Declare certain types external
        if n_defined == 0:
            remaining_names = {
                d_type.name for d_type in d_types if d_type.name not in already_defined
            }
            for d_type in d_types:
                only_external_undefined_fields = True
                for f in d_type.fields:
                    core_field_type = _core_type(f.python_type)
                    if (
                        core_field_type not in already_defined
                        and core_field_type in remaining_names
                    ):
                        only_external_undefined_fields = False
                        break
                if only_external_undefined_fields:
                    for f in d_type.fields:
                        core_field_type = _core_type(f.python_type)
                        if core_field_type not in already_defined:
                            lines.extend(["", ""])
                            lines.append(
                                f"from {default_package} import {core_field_type}"
                            )
                            already_defined.append(core_field_type)
                            n_defined += 1
                    break

    if not_defined:
        not_defined_list = "; ".join(
            [
                t.name
                + " ("
                + ", ".join(
                    _core_type(f.python_type)
                    for f in t.fields
                    if _core_type(f.python_type) not in already_defined
                )
                + ")"
                for t in not_defined
            ]
        )
        raise RuntimeError(f"Failed to define data types: {not_defined_list}")

    lines.append("")
    return lines


def operations(ops: list[Operation]) -> list[str]:
    lines = ["class OperationID(str, Enum):"]
    for op in ops:
        lines.append(f'    {op.name} = "{op.id}"')
    lines.append("")
    lines.append("")

    lines.append("OPERATIONS: Dict[OperationID, Operation] = {")
    for op in ops:
        lines.append(f"    OperationID.{op.name}: Operation(")
        lines.append(f'        id="{op.id}",')
        lines.append(f'        path="{op.path}",')
        lines.append(f'        verb="{op.verb}",')
        lines.append(f"        request_body_type={op.request_body_type},")
        lines.append("        response_body_type={")
        for code, resp_body in op.response_body_type.items():
            lines.append(f"            {code}: {resp_body},")
        lines.append("        }")
        lines.append("    ),")
    lines.append("}")

    return lines


def api_version(spec: dict) -> list[str]:
    version = spec.get("info", {}).get("version", None)
    if version:
        result = [f'API_VERSION = "{version}"']
    else:
        result = ["API_VERSION = None"]
    name = spec.get("info", {}).get("title", "unknown")
    result.append(
        f'"""Version of {name} OpenAPI specification from which the objects in this package were generated."""'
    )
    return result



"""Data types and operations from Remote ID Display Data Observation 0.3.0 OpenAPI"""

# This file is autogenerated; do not modify manually!

from __future__ import annotations

from enum import Enum

from implicitdict import ImplicitDict

from uas_standards.utils import Operation

API_VERSION = "0.3.0"
"""Version of Remote ID Display Data Observation OpenAPI specification from which the objects in this package were generated."""


class UAType(str, Enum):
    """The UA Type can help infer performance, speed, and duration of flights, for example, a
    "fixed wing" can generally fly in a forward direction only (as compared to a multi-rotor).
    This can also help differentiate aircraft types without sharing operationally sensitive
    information like the make and model of a particular aircraft. Make and model are
    anticipated to become available during the Registration ID lookup process. UAS Type is also
    useful for correlating visual observation with data received. The types were formulated
    based on unique flight characteristics.

    `HybridLift` is a fixed wing aircraft that can take off vertically.  `Helicopter` includes multirotor.

    `VTOL` is equivalent to HybridLift, and kept for retro-compatibility with F3411-v19
    """

    NotDeclared = "NotDeclared"
    Aeroplane = "Aeroplane"
    Helicopter = "Helicopter"
    Gyroplane = "Gyroplane"
    VTOL = "VTOL"
    HybridLift = "HybridLift"
    Ornithopter = "Ornithopter"
    Glider = "Glider"
    Kite = "Kite"
    FreeBalloon = "FreeBalloon"
    CaptiveBalloon = "CaptiveBalloon"
    Airship = "Airship"
    FreeFallOrParachute = "FreeFallOrParachute"
    Rocket = "Rocket"
    TetheredPoweredAircraft = "TetheredPoweredAircraft"
    GroundObstacle = "GroundObstacle"
    Other = "Other"


TimestampAccuracy = float
"""Declaration of timestamp accuracy, which is the largest difference between Timestamp and true time of applicability for any of the following fields: Latitude, Longitude, Geodetic Altitude, Pressure Altitude of Position, Height. to determine time of applicability of the location data provided.  Expressed in seconds, precise to 1/10ths of seconds. The accuracy reflects the 95% uncertainty bound value for the timestamp."""


class RIDOperationalStatus(str, Enum):
    """Indicates operational status of associated aircraft."""

    Undeclared = "Undeclared"
    Ground = "Ground"
    Airborne = "Airborne"
    Emergency = "Emergency"
    RemoteIDSystemFailure = "RemoteIDSystemFailure"


RIDTrack = float
"""Direction of flight expressed as a "True North-based" ground track angle.  This value is provided in degrees East of North with a minimum resolution of 1 degree."""


RIDSpeed = float
"""Ground speed of flight in meters per second."""


class SpeedAccuracy(str, Enum):
    """Provides quality/containment on horizontal ground speed.

    `SAUnknown`: Unknown speed accuracy

    `SA10mpsPlus`: > 10 m/s

    `SA10mps`: < 10 m/s

    `SA3mps`: < 3 m/s

    `SA1mps`: < 1 m/s

    `SA03mps`: < 0.3 m/s
    """

    SAUnknown = "SAUnknown"
    SA10mpsPlus = "SA10mpsPlus"
    SA10mps = "SA10mps"
    SA3mps = "SA3mps"
    SA1mps = "SA1mps"
    SA03mps = "SA03mps"


VerticalSpeed = float
"""Speed up (vertically) WGS84-HAE, m/s"""


Longitude = float
"""Degrees of longitude east of the Prime Meridian, with reference to the WGS84 ellipsoid."""


Latitude = float
"""Degrees of latitude north of the equator, with reference to the WGS84 ellipsoid."""


class OperatorAltitudeAltitudeType(str, Enum):
    """Source of data for the altitude field."""

    Takeoff = "Takeoff"
    Dynamic = "Dynamic"
    Fixed = "Fixed"


Altitude = float
"""An altitude, in meters, above the WGS84 ellipsoid."""


class UAClassificationEUCategory(str, Enum):
    EUCategoryUndefined = "EUCategoryUndefined"
    Open = "Open"
    Specific = "Specific"
    Certified = "Certified"


class UAClassificationEUClass(str, Enum):
    EUClassUndefined = "EUClassUndefined"
    Class0 = "Class0"
    Class1 = "Class1"
    Class2 = "Class2"
    Class3 = "Class3"
    Class4 = "Class4"
    Class5 = "Class5"
    Class6 = "Class6"


UAClassificationEU = dict
"""Expected keys:
* category
* class
"""


class HorizontalAccuracy(str, Enum):
    """This is based on ADS-B NACp (plus the one extra increment).

    `HAUnknown`: Unknown horizontal accuracy

    `HA10NMPlus`: > 10NM (18.52km)

    `HA10NM`: < 10NM (18.52km)

    `HA4NM`: < 4NM (7.408km)

    `HA2NM`: < 2NM (3.704km)

    `HA1NM`: < 1NM (1852m)

    `HA05NM`: < 0.5NM (926m)

    `HA03NM`: < 0.3NM (555.6m)

    `HA01NM`: < 0.1NM (185.2m)

    `HA005NM`: < 0.05NM (92.6m)

    `HA30m`: < 30m

    `HA10m`: < 10m

    `HA3m`: < 3m

    `HA1m`: < 1m
    """

    HAUnknown = "HAUnknown"
    HA10NMPlus = "HA10NMPlus"
    HA10NM = "HA10NM"
    HA4NM = "HA4NM"
    HA2NM = "HA2NM"
    HA1NM = "HA1NM"
    HA05NM = "HA05NM"
    HA03NM = "HA03NM"
    HA01NM = "HA01NM"
    HA005NM = "HA005NM"
    HA30m = "HA30m"
    HA10m = "HA10m"
    HA3m = "HA3m"
    HA1m = "HA1m"


class VerticalAccuracy(str, Enum):
    """This is based on ADS-B Geodetic Vertical Accuracy (GVA) (plus the three extra increments)

    `VAUnknown`: Unknown vertical accuracy

    `VA150mPlus`: > 150m

    `VA150m`: < 150m

    `VA45m`: < 45m

    `VA25m`: < 25m

    `VA10m`: < 10m

    `VA3m`: < 3m

    `VA1m`: < 1m
    """

    VAUnknown = "VAUnknown"
    VA150mPlus = "VA150mPlus"
    VA150m = "VA150m"
    VA45m = "VA45m"
    VA25m = "VA25m"
    VA10m = "VA10m"
    VA3m = "VA3m"
    VA1m = "VA1m"


class AltitudeReference(str, Enum):
    W84 = "W84"
    EGM96 = "EGM96"
    EGM2008 = "EGM2008"
    Other = "Other"


class RIDHeightReference(str, Enum):
    """The reference datum above which the height is reported."""

    TakeoffLocation = "TakeoffLocation"
    GroundLevel = "GroundLevel"


class RIDHeight(ImplicitDict):
    """A relative altitude for the purposes of remote ID."""

    distance: float
    """Distance above reference datum.  This value is provided in meters and must have a minimum resolution of 1 meter."""

    reference: RIDHeightReference
    """The reference datum above which the height is reported."""


class CurrentState(ImplicitDict):
    """Current state of an aircraft."""

    timestamp: str | None
    """Time at which the message containing the current state information was generated by the display provider."""

    timestamp_accuracy: TimestampAccuracy | None

    operational_status: RIDOperationalStatus | None

    track: RIDTrack | None

    speed: RIDSpeed | None

    speed_accuracy: SpeedAccuracy | None
    """Accuracy of horizontal ground speed."""

    vertical_speed: VerticalSpeed | None


class UAS(ImplicitDict):
    """UAS information"""

    id: str | None
    """Observed arbitrary and format agnostic (string) identification of a UAS."""

    eu_classification: UAClassificationEU | None
    """When this field is specified, the Classification Type is "European Union".  If no other classification field is specified, the Classification Type is "Undeclared"."""


class LatLngPoint(ImplicitDict):
    """Point on the earth's surface."""

    lng: Longitude

    lat: Latitude


class OperatorAltitude(ImplicitDict):
    """Altitude associated with the Remote Pilot"""

    altitude: Altitude | None

    altitude_type: OperatorAltitudeAltitudeType | None
    """Source of data for the altitude field."""


class MSLAltitude(ImplicitDict):
    meters: float | None = 0
    """Height, in meters, above a mean sea level datum."""

    reference_datum: AltitudeReference | None


class Operator(ImplicitDict):
    """Operator's information"""

    id: str | None
    """CAA-issued registration/license ID for the remote pilot or operator."""

    location: LatLngPoint | None

    altitude: OperatorAltitude | None


class Position(ImplicitDict):
    """A position on Earth."""

    lat: Latitude

    lng: Longitude

    alt: Altitude | None

    accuracy_h: HorizontalAccuracy | None
    """Horizontal error that is likely to be present in this reported position."""

    accuracy_v: VerticalAccuracy | None
    """Vertical error that is likely to be present in this reported position."""

    msl_alt: MSLAltitude | None

    height: RIDHeight | None


class Path(ImplicitDict):
    """Path followed by a flight."""

    positions: list[Position]
    """Sequential positions available for a flight."""


class GetDetailsResponse(ImplicitDict):
    """Response to a request to get details about a flight."""

    operator: Operator | None

    uas: UAS | None


class Flight(ImplicitDict):
    id: str
    """Identifier of flight that may be used to obtain details about the flight. This is not necessarily the UTM/flight ID in the remote ID system."""

    aircraft_type: UAType | None
    """Aircraft type of the flight."""

    current_state: CurrentState | None

    most_recent_position: Position | None
    """Most recent position known for the flight."""

    recent_paths: list[Path] | None
    """Paths the flight recently traveled, if available."""


class Cluster(ImplicitDict):
    """A general area containing one or more flight."""

    corners: list[Position]
    """Two opposite corners of a rectangular lat-lng box bounding the cluster."""

    area_sqm: float
    """Area of the cluster in square meters."""

    number_of_flights: int
    """Number of flights within the cluster."""


class GetDisplayDataResponse(ImplicitDict):
    """Response to a request for current data that would be visualized by a Display Application."""

    flights: list[Flight] | None = []
    """Current information for set of discovered flights whose precise locations are known."""

    clusters: list[Cluster] | None = []
    """Current information for sets of discovered flights whose precise locations are not known."""


class OperationID(str, Enum):
    GetDisplayData = "getDisplayData"
    GetDetails = "getDetails"


OPERATIONS: dict[OperationID, Operation] = {
    OperationID.GetDisplayData: Operation(
        id="getDisplayData",
        path="/display_data",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetDisplayDataResponse,
        },
    ),
    OperationID.GetDetails: Operation(
        id="getDetails",
        path="/display_data/{id}",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: GetDetailsResponse,
            404: None,
        },
    ),
}



# This tool generates Python data types and path constants from an OpenAPI YAML file.

import argparse
import os

import data_types
import flattening
import operations
import rendering
import yaml


def _parse_args():
    parser = argparse.ArgumentParser(
        description="Autogenerate Python data types from an OpenAPI YAML"
    )

    # Input/output specifications
    parser.add_argument(
        "--api", dest="api", type=str, help="Source YAML to preprocess."
    )
    parser.add_argument(
        "--python_output",
        dest="python_output",
        type=str,
        help="Output file for generated Python code",
    )
    parser.add_argument(
        "--default_package",
        dest="default_package",
        type=str,
        help="If this API refers to objects in another API, the Python package name where those other objects may be found",
        default="<not_defined>",
    )

    return parser.parse_args()


def main():
    args = _parse_args()

    # Parse OpenAPI
    with open(args.api) as f:
        spec = yaml.full_load(f)

    # Flatten external $refs
    flattening.flatten(spec, os.path.dirname(os.path.abspath(args.api)))

    # Parse data types
    types = data_types.parse(spec)

    # Parse operations
    ops = operations.get_operations(spec)

    # Render Python code
    with open(args.python_output, "w") as f:
        f.write(
            f'"""Data types and operations from {spec["info"]["title"]} {spec["info"]["version"]} OpenAPI"""\n\n'
        )
        f.write("\n".join(rendering.header(types)))
        f.write("\n\n\n")
        f.write("\n".join(rendering.api_version(spec)))
        f.write("\n".join(rendering.data_types(types, args.default_package)))
        f.write("\n\n")
        f.write("\n".join(rendering.operations(ops)))
        f.write("\n")


if __name__ == "__main__":
    main()



import dataclasses
from typing import Any


@dataclasses.dataclass
class ObjectField:
    """A data field within an Object data type"""

    api_name: str
    """Name of the field in the API (generally snake cased)"""

    python_type: str
    """The name of the Python data type which represents this field's value"""

    description: str
    """Documentation of this field"""

    required: bool
    """True if an instance of the parent object must specify a value for this field"""

    default: Any | None
    """Default value for field, if specified"""

    literal_default: bool = False
    """If true, render the default without quotes even if it's a string"""


@dataclasses.dataclass
class DataType:
    """A specific data type defined in the API"""

    name: str
    """Name of this data type, as defined in the API"""

    python_type: str = ""
    """Name of the Python data type ('ImplicitDict' for Object data types)"""

    description: str = ""
    """Documentation of this data type"""

    fields: list[ObjectField] = dataclasses.field(default_factory=list)
    """If this is an Object data type, a list of fields contained in that Object"""

    enum_values: dict[str, str] = dataclasses.field(default_factory=dict)
    """If this is a enum data type, a map from values it may take on to Python names for those values"""


python_primitives: dict[str, str] = {
    "string": "str",
    "boolean": "bool",
}
"""Maps OpenAPI `type` to Python primitive type"""

python_numbers: dict[str, str] = {
    "float": "float",
    "double": "float",
    "int32": "int",
    "int64": "int",
    "number": "float",
    "integer": "int",
}
"""Maps OpenAPI `format` (defaulting to `type` if `format` is missing) to Python primitive type"""


def is_primitive_python_type(python_type_name: str) -> bool:
    """True iff python_type_name describes a built-in Python type"""
    return (
        python_type_name in python_primitives.values()
        or python_type_name in python_numbers.values()
    )


def get_data_type_name(component_name: str, data_type_name: str) -> str:
    """Get the plain data type name from a $ref URI.

    :param component_name: $ref URI to the data type of interest
    :param data_type_name: context in which the data type is being retrieved (used for error message only)
    :return: Plain data type name in the relative $ref URI
    """
    if component_name == "":
        return ""
    elif component_name.startswith("#/components/schemas/"):
        return component_name[len("#/components/schemas/") :]
    else:
        if "#/components/schemas/" not in component_name:
            raise ValueError(
                f"$ref expected to contain `#/components/schemas/`, but found `{component_name}` instead for {data_type_name}"
            )
        name = get_data_type_name(
            component_name[component_name.index("#") :], data_type_name
        )
        print(
            f'WARNING: Assuming the variable type of {component_name} should be "{name}" and that it will be manually declared'
        )
        return name


def _parse_referenced_type_name(schema: dict, data_type_name: str) -> str:
    options = schema["anyOf"] if "anyOf" in schema else schema["allOf"]
    if len(options) != 1:
        raise NotImplementedError(
            f"Only one $ref is supported for anyOf and allOf; found {len(options)} elements instead"
        )
    option = options[0]
    if not isinstance(option, dict):
        raise ValueError(
            f"Expected dict entries in anyOf/allOf block; found {option} instead"
        )
    if len(option) != 1 or "$ref" not in option:
        raise NotImplementedError(
            f"The only element in anyOf/allOf must be a $ref dictionary; found {option} instead"
        )
    return get_data_type_name(option["$ref"], data_type_name)


def _snake_to_pascal(snake_case: str) -> str:
    words = snake_case.split("_")
    return "".join(w[0].upper() + w[1:] for w in words)


def make_object_field(
    python_object_name: str, api_field_name: str, schema: dict, required: set[str]
) -> tuple[ObjectField, list[DataType]]:
    """Parse a single field in a data type or endpoint parameter schema.

    :param python_object_name: Name of the Python object containing this field, for error messages and inline type names
    :param api_field_name: Name of the object field being parsed, according to the API
    :param schema: Definition of the object field being parsed
    :param required: The set of required fields for the parent object
    :return: Tuple of
      * The object field defined by the provided schema
      * Any additional data types incidentally defined in the provided schema
    """
    is_required = api_field_name in required
    default_value = schema["default"] if "default" in schema else None
    if "$ref" in schema:
        return (
            ObjectField(
                api_name=api_field_name,
                python_type=get_data_type_name(schema["$ref"], python_object_name),
                description=schema.get("description", ""),
                required=is_required,
                default=default_value,
            ),
            [],
        )
    elif "anyOf" in schema or "allOf" in schema:
        return (
            ObjectField(
                api_name=api_field_name,
                python_type=_parse_referenced_type_name(
                    schema, python_object_name + "." + api_field_name
                ),
                description=schema.get("description", ""),
                required=is_required,
                default=default_value,
            ),
            [],
        )
    else:
        type_name = python_object_name + _snake_to_pascal(api_field_name)
        data_type, additional_types = make_data_types(type_name, schema)
        if (
            is_primitive_python_type(data_type.python_type)
            and not data_type.enum_values
        ):
            # No additional type declaration needed
            if additional_types:
                raise RuntimeError(
                    f"{python_object_name} field type `{api_field_name}` was parsed as primitive {data_type.python_type} but also generated {len(additional_types)} additional types"
                )
            field_data_type = data_type.python_type
        elif data_type.python_type == "StringBasedDateTime":
            # No additional type declaration needed
            field_data_type = data_type.python_type
        elif data_type.python_type.startswith("List["):
            # Use array data type as-is
            field_data_type = data_type.python_type
        else:
            additional_types.append(data_type)
            field_data_type = data_type.name
        if len(data_type.enum_values) == 1 and default_value is None:
            default_value = (
                data_type.name + "." + next(iter(data_type.enum_values.values()))
            )
            literal_default = True
        else:
            literal_default = False
        return (
            ObjectField(
                api_name=api_field_name,
                python_type=field_data_type,
                description=data_type.description,
                required=is_required,
                default=default_value,
                literal_default=literal_default,
            ),
            additional_types,
        )


def _make_object_fields(
    python_object_name: str, properties: dict, required: set[str]
) -> tuple[list[ObjectField], list[DataType]]:
    fields: list[ObjectField] = []
    additional_types: list[DataType] = []
    for field_name, schema in properties.items():
        field, further_types = make_object_field(
            python_object_name, field_name, schema, required
        )
        additional_types.extend(further_types)
        fields.append(field)
    return fields, additional_types


def _make_python_enums(values: list[str]) -> dict[str, str]:
    valid_characters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"
    valid_start_characters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_"
    result = {}
    for value in values:
        name = str(value)
        if name[0] not in valid_start_characters:
            name = "_" + name
        name = "".join(c if c in valid_characters else "_" for c in name)
        result[str(value)] = name
    return result


def make_data_types(api_name: str, schema: dict) -> tuple[DataType, list[DataType]]:
    """Parse all data types necessary to express the provided data type schema.

    In addition to the primary data type described by `name`, this routine also
    generates additional data types defined inline in the provided schema.

    :param api_name: Name of the primary data type being parsed, according to the API
    :param schema: Definition of the data type being parsed
    :return: Tuple of
      * The primary data defined by the provided schema
      * Any additional data types incidentally defined in the provided schema
    """
    data_type = DataType(name=api_name)
    additional_types = []

    if "description" in schema:
        data_type.description = schema["description"]

    if "type" in schema:
        if schema["type"] in python_primitives:
            data_type.python_type = python_primitives[schema["type"]]
            if (
                data_type.python_type == "str"
                and "format" in schema
                and schema["format"] == "date-time"
            ):
                data_type.python_type = "StringBasedDateTime"
        elif schema["type"] in {"number", "integer"}:
            data_type.python_type = python_numbers.get(
                schema.get("format", schema["type"]), ""
            )
            if not data_type.python_type:
                raise ValueError(
                    "Unrecognized numeric format `{}` for {}".format(
                        schema.get("format", "<missing>"), api_name
                    )
                )
        elif schema["type"] == "array":
            if "items" in schema:
                items = schema["items"]
                if "$ref" in items:
                    item_type_name = get_data_type_name(items["$ref"], api_name)
                else:
                    item_type, further_types = make_data_types(api_name + "Item", items)
                    additional_types.extend(further_types)
                    if item_type.description != "" or not is_primitive_python_type(
                        item_type.python_type
                    ):
                        additional_types.append(item_type)
                        item_type_name = item_type.name
                    else:
                        item_type_name = item_type.python_type
                data_type.python_type = f"List[{item_type_name}]"
            else:
                raise ValueError(
                    f"Missing `items` declaration for {api_name} array type"
                )
        elif schema["type"] == "object":
            data_type.python_type = "ImplicitDict"
            data_type.fields, further_types = _make_object_fields(
                api_name, schema.get("properties", {}), set(schema.get("required", []))
            )
            additional_types.extend(further_types)
        else:
            raise ValueError(
                "Unrecognized type `{}` in {} type".format(schema["type"], api_name)
            )
    elif "anyOf" in schema or "allOf" in schema:
        data_type.python_type = _parse_referenced_type_name(schema, api_name)

    if "enum" in schema:
        data_type.enum_values = _make_python_enums(schema["enum"])

    if not data_type.python_type:
        data_type.python_type = "str"

    return data_type, additional_types


def parse(spec: dict) -> list[DataType]:
    if "components" not in spec:
        raise ValueError("Missing `components` in YAML")
    components = spec["components"]
    if "schemas" not in components:
        raise ValueError("Missing `schemas` in `components`")
    declared_types = []
    for name, schema in components["schemas"].items():
        data_type, additional_types = make_data_types(name, schema)
        declared_types.extend(additional_types)
        declared_types.append(data_type)
    return declared_types



from __future__ import annotations

from enum import Enum
from typing import Any, Literal

from implicitdict import ImplicitDict, StringBasedDateTime


class CodeAuthorityRole(str, Enum):
    AUTHORIZATION = "AUTHORIZATION"
    NOTIFICATION = "NOTIFICATION"
    INFORMATION = "INFORMATION"


class CodeDaylightEventType(str, Enum):
    BMCT = "BMCT"
    SR = "SR"
    SS = "SS"
    EECT = "EECT"


class UomDistance(str, Enum):
    M = "m"
    FT = "ft"


CodeZoneIdentifierType = str

CodeCountryISOType = str


class CodeZoneVariantType(str, Enum):
    COMMON = "COMMON"
    CUSTOMIZED = "CUSTOMIZED"


class CodeZoneType(str, Enum):
    USPACE = "USPACE"
    PROHIBITED = "PROHIBITED"
    REQ_AUTHORIZATION = "REQ_AUTHORIZATION"
    CONDITIONAL = "CONDITIONAL"
    NO_RESTRICTION = "NO_RESTRICTION"


ConditionExpressionType = str


class CodeZoneReasonType(str, Enum):
    AIR_TRAFFIC = "AIR_TRAFFIC"
    SENSITIVE = "SENSITIVE"
    PRIVACY = "PRIVACY"
    POPULATION = "POPULATION"
    NATURE = "NATURE"
    NOISE = "NOISE"
    EMERGENCY = "EMERGENCY"
    DAR = "DAR"
    OTHER = "OTHER"


class TextShortType(ImplicitDict):
    # This complies with the JSON schema provided in the appendix E of the standard
    # Though, from the description, lang should be optional and text required.
    text: str | None
    lang: str


class TextLongType(ImplicitDict):
    # This complies with the JSON schema provided in the appendix E of the standard
    # Though, from the description, lang should be optional and text required.
    text: str | None
    lang: str


class CodeWeekDayType(str, Enum):
    MON = "MON"
    TUE = "TUE"
    WED = "WED"
    THU = "THU"
    FRI = "FRI"
    SAT = "SAT"
    SUN = "SUN"
    ANY = "ANY"


DateTimeType = StringBasedDateTime

TimeInterval = str  # TODO: Create appropriate type

TimeType = str  # TODO: Create appropriate type


class CodeYesNoType(str, Enum):
    YES = "YES"
    NO = "NO"


URNType = str


class VerticalLayer(ImplicitDict):
    upper: float
    upperReference: CodeVerticalReferenceType
    lower: float
    lowerReference: CodeVerticalReferenceType
    uom: UomDistance


class Layered(ImplicitDict):
    layer: VerticalLayer | None
    bbox: list[float] | None


class Point(Layered):
    type: Literal["Point"]
    coordinates: list[float]
    extent: ExtentCircle | None


class ExtentCircle(ImplicitDict):
    subType: Literal["Circle"]
    radius: float


class LineString(Layered):
    type: Literal["LineString"]
    coordinates: list[list[float]]


class Polygon(Layered):
    type: Literal["Polygon"]
    coordinates: list[list[list[float]]]


class MultiPoint(Layered):
    type: Literal["MultiPoint"]
    coordinates: list[list[float]]


class MultiLineString(Layered):
    type: Literal["MultiLineString"]
    coordinates: list[list[list[float]]]


class MultiPolygon(Layered):
    type: Literal["MultiPolygon"]
    coordinates: list[list[list[list[float]]]]


class GeometryCollection(ImplicitDict):
    type: Literal["GeometryCollection"]
    geometries: list[Any]


Geometry = (
    Point
    | LineString
    | Polygon
    | MultiPoint
    | MultiLineString
    | MultiPolygon
    | GeometryCollection
    | dict[str, Any]
)  # fallback for GeometryCollection or future types


class CodeVerticalReferenceType(str, Enum):
    AGL = "AGL"
    AMSL = "AMSL"
    WGS84 = "WGS84"


# Data models objects


class DatasetMetadata(ImplicitDict):
    provider: list[TextShortType] | None
    issued: DateTimeType | None
    validFrom: DateTimeType | None
    validTo: DateTimeType | None
    description: list[TextShortType] | None
    otherGeoid: URNType | None
    technicalLimitations: list[TextShortType] | None


class UASZone(ImplicitDict):
    identifier: CodeZoneIdentifierType
    country: CodeCountryISOType
    name: list[TextShortType] | None
    type: CodeZoneType
    variant: CodeZoneVariantType
    restrictionConditions: ConditionExpressionType | None
    region: int | None
    reason: list[CodeZoneReasonType] | None
    otherReasonInfo: list[TextShortType] | None
    regulationExemption: CodeYesNoType | None
    message: list[TextLongType] | None
    extendedProperties: dict[str, Any] | None
    limitedApplicability: list[TimePeriod] | None
    zoneAuthority: list[Authority]
    dataSource: Metadata | None


class TimePeriod(ImplicitDict):
    startDateTime: DateTimeType | None
    endDateTime: DateTimeType | None
    schedule: list[DailyPeriod] | None


class DailyPeriod(ImplicitDict):
    day: list[CodeWeekDayType]
    startTime: TimeType | None
    startEvent: CodeDaylightEventType | None
    endTime: TimeType | None
    endEvent: CodeDaylightEventType | None


class Authority(ImplicitDict):
    purpose: CodeAuthorityRole
    intervalBefore: TimeInterval | None
    name: list[TextShortType] | None
    service: list[TextShortType] | None
    contactName: list[TextShortType] | None
    siteURL: TextShortType | None
    email: TextShortType | None
    phone: TextShortType | None


class Metadata(ImplicitDict):
    creationDateTime: DateTimeType | None
    updateDateTime: DateTimeType | None
    originator: str | None


class Feature(ImplicitDict):
    type: Literal["Feature"]
    id: int | str | None
    properties: UASZone | None
    geometry: Geometry | None
    bbox: list[float] | None


class ED318Schema(ImplicitDict):
    """Top-level ED-318 FeatureCollection payload."""

    type: Literal["FeatureCollection"]
    name: str | None
    metadata: DatasetMetadata
    title: str | None
    bbox: list[float] | None
    features: list[Feature]

    @staticmethod
    def from_dict(obj: dict) -> ED318Schema:
        """Parse a raw dict into typed ED-318 classes using implicitdict."""
        return ImplicitDict.parse(obj, ED318Schema)



from . import flight_planning, geo_awareness, geospatial_map, rid, scd, versioning

__all__ = [
    "flight_planning",
    "geospatial_map",
    "geo_awareness",
    "rid",
    "scd",
    "versioning",
]



from . import ansi_cta_2063_a, astm, en4709_02, eurocae_ed269, interuss
from .utils import Operation

__all__ = [
    "astm",
    "interuss",
    "Operation",
    "ansi_cta_2063_a",
    "en4709_02",
    "eurocae_ed269",
]



from dataclasses import dataclass


@dataclass
class Operation:
    id: str
    path: str
    verb: str
    request_body_type: str | None
    response_body_type: dict[int, str | None]

    @property
    def name(self) -> str:
        return self.id[0].upper() + self.id[1:]


def get_operations(spec: dict) -> list[Operation]:
    ref_prefix = "#/components/schemas/"
    operations: list[Operation] = []
    all_verbs = {"get", "put", "post", "delete", "patch"}
    for path, body in spec["paths"].items():
        for k, v in body.items():
            if k not in all_verbs:
                continue

            ref = (
                v.get("requestBody", {})
                .get("content", {})
                .get("application/json", {})
                .get("schema", {})
                .get("$ref", "")
            )
            request_body_type = (
                ref[len(ref_prefix) :] if ref.startswith(ref_prefix) else None
            )

            response_body_type = {}
            for code, response_content in v["responses"].items():
                ref = (
                    response_content.get("content", {})
                    .get("application/json", {})
                    .get("schema", {})
                    .get("$ref", "")
                )
                response_body_type[int(code)] = (
                    ref[len(ref_prefix) :] if ref.startswith(ref_prefix) else None
                )

            op_id = v.get("operationId", None)
            if not op_id:
                print(f"WARNING: {k} {path} is missing an operationId")
                path_parts = path.split("/")
                if path_parts[-1].startswith("{") or "_" in path_parts[-1]:
                    raise NotImplementedError("Cannot construct operationId")
                op_id = k.lower() + path_parts[-1][0].upper() + path_parts[-1][1:]
            operations.append(
                Operation(
                    id=op_id,
                    path=path,
                    verb=k.upper(),
                    request_body_type=request_body_type,
                    response_body_type=response_body_type,
                )
            )
    return operations



from dataclasses import dataclass


@dataclass
class Operation:
    id: str
    """Operation ID of this operation."""

    path: str
    """Path relative to the base URL at which the operation is accessed."""

    verb: str
    """HTTP verb used to invoke the operation."""

    request_body_type: type | None
    """Data type describing the contents of the request body, or None if no request body."""

    response_body_type: dict[int, type | None]
    """Data type describing the contents of the response body provided with the corresponding status code."""


__all__ = ["Operation"]



"""Data types and operations from Remote ID Test Data Injection 0.5.1 OpenAPI"""

# This file is autogenerated; do not modify manually!

from __future__ import annotations

from enum import Enum

from implicitdict import ImplicitDict, StringBasedDateTime

from uas_standards.utils import Operation

API_VERSION = "0.5.1"
"""Version of Remote ID Test Data Injection OpenAPI specification from which the objects in this package were generated."""

RIDFlightID = str
"""ID, unique to a remote ID service provider, which identifies a particular flight for which the remote ID service provider is providing remote ID services.

The following characters are not allowed: \\0 \\t \\r \\n # % / : ? @ [ \\ ]
"""


class RIDAuthData(ImplicitDict):
    """Additional authentication data."""

    format: int | None = 0
    """Format of additional authentication data.

    0: None
    1: UAS ID Signature
    2: Operator ID Signature
    3: Message Set Signature
    4: Authentication Provided by Network Remote ID
    5: Specific Method
    6-9: Reserved for Spec
    10-15: Available for Private Use
    """

    data: str | None = ""
    """Authentication data in form specified by `format`."""


SpecificSessionID = str
"""A unique 20 byte ID intended to identify a specific flight (session) while providing a
greater level of privacy to the operator. The first byte is the registered unique Specific Session ID
Type maintained by a registrar (see Annex A5), and the remaining 19 bytes is the Session ID per the
Specific Session ID Type specification.

Initial scheme registry entries include:

0 – reserved
1 – Internet Engineering Task Force (IETF) Drone Remote Identification Protocol (DRIP) entity ID
2 – IEEE 1609.2 HashedID8

Expressed as a hexadecimal string of the underlying data bytes.  A recipient should ignore any dashes in this value.  If fewer than 20 bytes are specified, the remaining bytes at the end are assumed to have value 0.
"""


class UASID(ImplicitDict):
    """Identification of the UAS performing this flight.  At least one field of this object must be specified."""

    serial_number: str | None = ""
    """This is expressed in the CTA-2063-A Serial Number format."""

    registration_id: str | None = ""
    """If a CAA provides a method of registering UAS, this number is provided by the CAA or its authorized representative.  Format is <ICAO Nationality MarkA>.<CAA Assigned ID>, ASCII encoded, only uppercase letters (A-Z), dot (.), and digits (0-9) are allowed. Example is valid for the US."""

    utm_id: str | None = ""
    """A UTM-provided universally unique ID traceable to a non-obfuscated ID that acts as a "session id" to protect exposure of operationally sensitive information."""

    specific_session_id: SpecificSessionID | None


class TimeFormat(str, Enum):
    RFC3339 = "RFC3339"


class Time(ImplicitDict):
    value: StringBasedDateTime
    """RFC3339-formatted time/date string.  The time zone must be 'Z'."""

    format: TimeFormat = TimeFormat.RFC3339


class UserNotification(ImplicitDict):
    """Notification observed by virtual user."""

    observed_at: Time
    """Time at which the virtual user observed the notification."""

    message: str | None
    """Message presented to the user, description of notification, or other means of helping identify the nature of the notification, for the purpose of increased readability of test reports."""


class QueryUserNotificationsResponse(ImplicitDict):
    """Response object for query request for notifications observed by the virtual user."""

    user_notifications: list[UserNotification]
    """List of applicable observed user notifications."""


Latitude = float
"""Degrees of latitude north of the equator, with reference to the WGS84 ellipsoid."""


Longitude = float
"""Degrees of longitude east of the Prime Meridian, with reference to the WGS84 ellipsoid."""


Altitude = float
"""An altitude, in meters, above the WGS84 ellipsoid."""


class RIDHeightReference(str, Enum):
    """The reference datum above which the height is reported."""

    TakeoffLocation = "TakeoffLocation"
    GroundLevel = "GroundLevel"


class RIDHeight(ImplicitDict):
    """A relative altitude for the purposes of remote ID."""

    distance: float
    """Distance above reference datum.  This value is provided in meters and must have a minimum resolution of 1 meter."""

    reference: RIDHeightReference
    """The reference datum above which the height is reported."""


class HorizontalAccuracy(str, Enum):
    """This is based on ADS-B NACp (plus the one extra increment).

    `HAUnknown`: Unknown horizontal accuracy

    `HA10NMPlus`: > 10NM (18.52km)

    `HA10NM`: < 10NM (18.52km)

    `HA4NM`: < 4NM (7.408km)

    `HA2NM`: < 2NM (3.704km)

    `HA1NM`: < 1NM (1852m)

    `HA05NM`: < 0.5NM (926m)

    `HA03NM`: < 0.3NM (555.6m)

    `HA01NM`: < 0.1NM (185.2m)

    `HA005NM`: < 0.05NM (92.6m)

    `HA30m`: < 30m

    `HA10m`: < 10m

    `HA3m`: < 3m

    `HA1m`: < 1m
    """

    HAUnknown = "HAUnknown"
    HA10NMPlus = "HA10NMPlus"
    HA10NM = "HA10NM"
    HA4NM = "HA4NM"
    HA2NM = "HA2NM"
    HA1NM = "HA1NM"
    HA05NM = "HA05NM"
    HA03NM = "HA03NM"
    HA01NM = "HA01NM"
    HA005NM = "HA005NM"
    HA30m = "HA30m"
    HA10m = "HA10m"
    HA3m = "HA3m"
    HA1m = "HA1m"


class VerticalAccuracy(str, Enum):
    """This is based on ADS-B Geodetic Vertical Accuracy (GVA) (plus the three extra increments)

    `VAUnknown`: Unknown vertical accuracy

    `VA150mPlus`: > 150m

    `VA150m`: < 150m

    `VA45m`: < 45m

    `VA25m`: < 25m

    `VA10m`: < 10m

    `VA3m`: < 3m

    `VA1m`: < 1m
    """

    VAUnknown = "VAUnknown"
    VA150mPlus = "VA150mPlus"
    VA150m = "VA150m"
    VA45m = "VA45m"
    VA25m = "VA25m"
    VA10m = "VA10m"
    VA3m = "VA3m"
    VA1m = "VA1m"


OperatorId = str
"""CAA-issued registration/license ID for the remote pilot or operator. """


class LatLngPoint(ImplicitDict):
    """Point on the earth's surface."""

    lng: Longitude

    lat: Latitude


class OperatorAltitudeAltitudeType(str, Enum):
    """Source of data for the altitude field."""

    Takeoff = "Takeoff"
    Dynamic = "Dynamic"
    Fixed = "Fixed"


class OperatorAltitude(ImplicitDict):
    """Altitude associated with the Remote Pilot"""

    altitude: Altitude | None

    altitude_type: OperatorAltitudeAltitudeType | None
    """Source of data for the altitude field."""


class UAClassificationEUCategory(str, Enum):
    EUCategoryUndefined = "EUCategoryUndefined"
    Open = "Open"
    Specific = "Specific"
    Certified = "Certified"


class UAClassificationEUClass(str, Enum):
    EUClassUndefined = "EUClassUndefined"
    Class0 = "Class0"
    Class1 = "Class1"
    Class2 = "Class2"
    Class3 = "Class3"
    Class4 = "Class4"
    Class5 = "Class5"
    Class6 = "Class6"


UAClassificationEU = dict
"""Expected keys:
* category
* class
"""


TimestampAccuracy = float
"""Declaration of timestamp accuracy, which is the largest difference between Timestamp and true time of applicability for any of the following fields: Latitude, Longitude, Geodetic Altitude, Pressure Altitude of Position, Height. to determine time of applicability of the location data provided.  Expressed in seconds, precise to 1/10ths of seconds. The accuracy reflects the 95% uncertainty bound value for the timestamp."""


class RIDOperationalStatus(str, Enum):
    """Indicates operational status of associated aircraft."""

    Undeclared = "Undeclared"
    Ground = "Ground"
    Airborne = "Airborne"
    Emergency = "Emergency"
    RemoteIDSystemFailure = "RemoteIDSystemFailure"


RIDTrack = float
"""Direction of flight expressed as a "True North-based" ground track angle.  This value is provided in degrees East of North with a minimum resolution of 1 degree."""


RIDSpeed = float
"""Ground speed of flight in meters per second."""


class SpeedAccuracy(str, Enum):
    """Provides quality/containment on horizontal ground speed.

    `SAUnknown`: Unknown speed accuracy

    `SA10mpsPlus`: > 10 m/s

    `SA10mps`: < 10 m/s

    `SA3mps`: < 3 m/s

    `SA1mps`: < 1 m/s

    `SA03mps`: < 0.3 m/s
    """

    SAUnknown = "SAUnknown"
    SA10mpsPlus = "SA10mpsPlus"
    SA10mps = "SA10mps"
    SA3mps = "SA3mps"
    SA1mps = "SA1mps"
    SA03mps = "SA03mps"


VerticalSpeed = float
"""Speed up (vertically) WGS84-HAE, m/s"""


class UAType(str, Enum):
    """The UA Type can help infer performance, speed, and duration of flights, for example, a
    "fixed wing" can generally fly in a forward direction only (as compared to a multi-rotor).
    This can also help differentiate aircraft types without sharing operationally sensitive
    information like the make and model of a particular aircraft. Make and model are
    anticipated to become available during the Registration ID lookup process. UAS Type is also
    useful for correlating visual observation with data received. The types were formulated
    based on unique flight characteristics.

    `HybridLift` is a fixed wing aircraft that can take off vertically.  `Helicopter` includes multirotor.

    `VTOL` is equivalent to HybridLift, and kept for retro-compatibility with F3411-v19
    """

    NotDeclared = "NotDeclared"
    Aeroplane = "Aeroplane"
    Helicopter = "Helicopter"
    Gyroplane = "Gyroplane"
    VTOL = "VTOL"
    HybridLift = "HybridLift"
    Ornithopter = "Ornithopter"
    Glider = "Glider"
    Kite = "Kite"
    FreeBalloon = "FreeBalloon"
    CaptiveBalloon = "CaptiveBalloon"
    Airship = "Airship"
    FreeFallOrParachute = "FreeFallOrParachute"
    Rocket = "Rocket"
    TetheredPoweredAircraft = "TetheredPoweredAircraft"
    GroundObstacle = "GroundObstacle"
    Other = "Other"


class RIDAircraftPosition(ImplicitDict):
    """Position of an aircraft as reported for remote ID purposes.
    The following fields are required for a successful injection, but are not included in the 'required' field to allow the uss_qualifier to check SP behavior in situations where a networked UAS would not provide the necessary data to participate in Network Remote ID:
    - lat - lng - alt
    """

    lat: Latitude | None

    lng: Longitude | None

    alt: Altitude | None

    height: RIDHeight | None

    accuracy_h: HorizontalAccuracy | None
    """Horizontal error that is likely to be present in this reported position.  Required when `extrapolated` field is true and always in the entry for the current state."""

    accuracy_v: VerticalAccuracy | None
    """Vertical error that is likely to be present in this reported position.  Required when `extrapolated` field is true and always in the entry for the current state."""

    extrapolated: bool | None
    """True if this position was generated primarily by computation rather than primarily from a direct instrument measurement.  Assumed false if not specified."""

    pressure_altitude: float | None
    """The uncorrected altitude (based on reference standard 29.92 inHg, 1013.25 mb) provides a reference for algorithms that utilize "altitude deltas" between aircraft.  This value is provided in meters and must have a minimum resolution of 1 meter."""


class RIDFlightDetails(ImplicitDict):
    """Details about a flight reported by a remote ID service provider.  At least one of the registration or serial fields must be filled if required by CAA."""

    id: str
    """ID for this flight, matching argument in request."""

    operator_id: OperatorId | None

    operator_location: LatLngPoint | None
    """Location of party controlling the aircraft."""

    operation_description: str | None
    """Free-text field that enables the operator to describe the purpose of a flight, if so desired."""

    auth_data: RIDAuthData | None

    serial_number: str | None
    """This field is DEPRECATED and will be removed in a future version; use uas_id.serial_number instead. Can be specified when no registration ID exists and required by law in a region. This is expressed in the ANSI/CTA-2063-A Physical Serial Number format."""

    registration_number: str | None
    """This field is DEPRECATED and will be removed in a future version; use uas_id.registration_number instead. If a CAA provides a method of registering UAS, this number is provided by the CAA or its authorized representative.  Required when required by law in a region."""

    uas_id: UASID | None

    operator_altitude: OperatorAltitude | None

    eu_classification: UAClassificationEU | None
    """When this field is specified, the Classification Type is "European Union".  If no other classification field is specified, the Classification Type is "Undeclared"."""


class RIDRecentAircraftPosition(ImplicitDict):
    time: StringBasedDateTime
    """Time at which this position applied.  RFC 3339 format, per OpenAPI specification."""

    position: RIDAircraftPosition


class RIDAircraftState(ImplicitDict):
    """State of an aircraft for the purposes of remote ID.
    The following fields are required for a successful injection, but are not included in the 'required' field to allow the uss_qualifier to check SP behavior in situations where a networked UAS would not provide the necessary data to participate in Network Remote ID:
    - timestamp - timestamp_accuracy - accuracy_h - accuracy_v - speed - vertical_speed - track - speed_accuracy - position
    """

    timestamp: StringBasedDateTime | None
    """Time at which this state was valid.  This may be the time coming from the source, such as a GPS, or the time when the system computes the values using an algorithm such as an Extended Kalman Filter (EKF).  Timestamp must be expressed with a minimum resolution of 1/10th of a second.  RFC 3339 format, per OpenAPI specification."""

    timestamp_accuracy: TimestampAccuracy | None

    operational_status: RIDOperationalStatus | None

    position: RIDAircraftPosition | None

    track: RIDTrack | None

    speed: RIDSpeed | None

    speed_accuracy: SpeedAccuracy | None
    """Accuracy of horizontal ground speed."""

    vertical_speed: VerticalSpeed | None

    height: RIDHeight | None

    group_radius: float | None
    """Farthest horizontal distance from reported group location at which an aircraft in the group may be located (meters).  This value contains the "Operating Area Radius" data from the common data dictionary when group operation area is specified by point-radius."""

    group_ceiling: float | None
    """Maximum altitude (meters WGS84-HAE) of Group Operation.  This value contains the "Operating Area Ceiling" data from the common data dictionary when group operation area is specified by point-radius."""

    group_floor: float | None
    """Minimum altitude (meters WGS84-HAE) of Group Operation.  If not specified, ground level shall be assumed.  This value contains the "Operating Area Floor" data from the common data dictionary when group operation area is specified by point-radius."""

    group_count: int | None
    """When operating a group (or formation or swarm), number of aircraft in group.  This value contains the "Operating Area Count" data from the common data dictionary when group operation area is specified by point-radius."""

    group_time_start: StringBasedDateTime | None
    """Time at which a group operation starts.  This value contains the "Operation Area Start" data from the common data dictionary when group operation area is specified by point-radius."""

    group_time_end: StringBasedDateTime | None
    """Time at which a group operation starts.  This value contains the "Operation Area End" data from the common data dictionary when group operation area is specified by point-radius."""


class TestFlightDetails(ImplicitDict):
    """The set of data with which the Service Provider system under test should respond when queried for the details of a test flight."""

    effective_after: str
    """The time after which the Service Provider system under test should respond with `details`, unless other `details` with a more recent `effective_after` time (before the current time) are available."""

    details: RIDFlightDetails
    """The details of the flight. Follows the RIDFlightDetails schema from the ASTM remote ID standard."""


class TestFlight(ImplicitDict):
    """The set of data to be provided to a Service Provider's test data injector for a single flight."""

    injection_id: str
    """ID of the injected test flight.  Remains the same regardless of the flight ID/UTM ID reported in the system."""

    aircraft_type: UAType | None
    """Aircraft type of the injected test flight."""

    telemetry: list[RIDAircraftState]
    """The set of telemetry data that should be injected into the system for this flight. Each element follows the RIDAircraftState schema from the ASTM remote ID standard."""

    details_responses: list[TestFlightDetails]
    """The details of the flight as a function of time."""


class CreateTestParameters(ImplicitDict):
    """A complete set of data to be injected into a Service Provider system under test."""

    requested_flights: list[TestFlight]
    """One or more logical flights, each containing test data to inject into the system. Elements should be sorted in ascending order of `timestamp`."""


class ChangeTestResponse(ImplicitDict):
    injected_flights: list[TestFlight]
    """The complete set of test data actually injected into the Service Provider system under test."""

    version: str
    """Version of test.  Used to delete test."""


class DeleteTestResponse(ImplicitDict):
    injected_flights: list[TestFlight]
    """The complete set of test data deleted."""


class OperationID(str, Enum):
    CreateTest = "createTest"
    DeleteTest = "deleteTest"
    QueryUserNotifications = "QueryUserNotifications"


OPERATIONS: dict[OperationID, Operation] = {
    OperationID.CreateTest: Operation(
        id="createTest",
        path="/tests/{test_id}",
        verb="PUT",
        request_body_type=CreateTestParameters,
        response_body_type={
            200: ChangeTestResponse,
            409: None,
        },
    ),
    OperationID.DeleteTest: Operation(
        id="deleteTest",
        path="/tests/{test_id}/{version}",
        verb="DELETE",
        request_body_type=None,
        response_body_type={
            200: DeleteTestResponse,
        },
    ),
    OperationID.QueryUserNotifications: Operation(
        id="QueryUserNotifications",
        path="/user_notifications",
        verb="GET",
        request_body_type=None,
        response_body_type={
            200: QueryUserNotificationsResponse,
            400: None,
            401: None,
            403: None,
        },
    ),
}
